// This file is @generated by prost-build.
#[derive(Clone, Copy, PartialEq, Eq, Hash, ::prost::Message)]
pub struct SmokeRequest {
}
#[derive(Clone, Copy, PartialEq, Eq, Hash, ::prost::Message)]
pub struct SmokeResponse {
}
/// Just a smoke test service.
trait SmokeService {
    // A detached comment block.

    /// Blow some smoke.
    fn blow_smoke(_: SmokeRequest) -> SmokeResponse;
}
pub mod utils { }
